from eth_typing import Address

OrderBookContractAddress = Address("0x0300000000000000000000000000000000000000")
CHAIN_ID = 321123
MAX_GAS_LIMIT = 7_000_000  # 7 million
GAS_PER_ORDER = 200_000  # 200k
